@@ -1,6 +1,6 @@
 /*
  * UnixCW - Unix CW (Morse code) training program
- * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
+ * Copyright (C) 2001, 2002  Simon Baldwin (simonb@caldera.com)
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -17,7 +17,7 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *
  *
- * cw.c - Filter program to output morse code on the system console.
+ * cw.c - Filter program to generate morse code.
  *
  */
 
@@ -56,8 +56,8 @@
 #include "cw.h"
 
 /* Macro definitions. */
-#define	VERSION		"cw version 2.0"
-#define	COPYRIGHT	"Copyright (C) 2001  Simon Baldwin\n"		\
+#define	VERSION		"cw version 2.2"
+#define	COPYRIGHT	"Copyright (C) 2001, 2002  Simon Baldwin\n"		\
 "This program comes with ABSOLUTELY NO WARRANTY; for details\n"		\
 "please see the file 'COPYING' supplied with the source code.\n"	\
 "This is free software, and you are welcome to redistribute it\n"	\
@@ -131,23 +131,31 @@ print_help ()
 {
 	int	min_speed, max_speed;		/* Cw library speed limits */
 	int	min_frequency, max_frequency;	/* Cw library frequncy limits */
+	int	min_volume, max_volume;		/* Cw library volume limits */
 	int	min_gap, max_gap;		/* Cw library gap limits */
 	assert (argv0 != NULL);
 
 	/* Read the cw library limits on its operating parameters. */
 	cw_get_speed_limits	(&min_speed, &max_speed);
 	cw_get_frequency_limits	(&min_frequency, &max_frequency);
+	cw_get_volume_limits	(&min_volume, &max_volume);
 	cw_get_gap_limits	(&min_gap, &max_gap);
 
 	printf (
 #if defined(HAVE_GETOPT_LONG)
 	"Usage: %s [options...]\n\n\
-	-d, --device=DEVICE	use DEVICE for sound ioctl [default stdout]\n\
+	-s, --sound=SOURCE	generate sound on SOURCE [default 'soundcard']\n\
+				one of 's[oundcard]', 'c[onsole]', or 'b[oth]'\n\
+	-x, --sdevice=SDEVICE	use SDEVICE for soundcard [default %s]\n\
+	-y, --mdevice=MDEVICE	use MDEVICE for sound mixer [default %s]\n\
+	-d, --cdevice=CDEVICE	use CDEVICE for sound ioctl [default %s]\n\
 	-f, --file=FILE		read from FILE [default stdin]\n\
 	-w, --wpm=WPM		set initial words per minute [default %d]\n\
 				valid WPM values are between %d and %d\n\
 	-t, --hz,--tone=HZ	set initial tone to HZ [default %d]\n\
 				valid HZ values are between %d and %d\n\
+	-v, --volume=PERCENT	set initial volume to PERCENT [default %d]\n\
+				valid PERCENT values are between %d and %d\n\
 	-g, --gap=GAP		set extra gap between letters [default %d]\n\
 				valid GAP values are between %d and %d\n\
 	-e, --noecho		don't echo sending to stdout [default echo]\n\
@@ -156,15 +164,21 @@ print_help ()
 	-o, --nocombinations	don't allow [...] combinations [default combo]\n\
 	-p, --nocomments	don't allow {...} comments [default comments]\n\
 	-h, --help		print this message\n\
-	-v, --version		output version information and exit\n\n",
+	-V, --version		output version information and exit\n\n",
 #else  /* not HAVE_GETOPT_LONG */
 	"Usage: %s [options...]\n\n\
-	-d DEVICE	use DEVICE for sound ioctl [default stdout]\n\
+	-s SOURCE	generate sound on SOURCE [default 'console']\n\
+			one of 'c[onsole]', 's[oundcard]', or 'b[oth]'\n\
+	-x SDEVICE	use SDEVICE for soundcard [default %s]\n\
+	-y MDEVICE	use MDEVICE for sound mixer [default %s]\n\
+	-d CDEVICE	use CDEVICE for sound ioctl [default %s]\n\
 	-f FILE		read from FILE [default stdin]\n\
 	-w WPM		set initial words per minute [default %d]\n\
 			valid WPM values are between %d and %d\n\
 	-t HZ		set initial tone to HZ [default %d]\n\
 			valid HZ values are between %d and %d\n\
+	-v PERCENT	set initial volume to PERCENT [default %d]\n\
+			valid PERCENT values are between %d and %d\n\
 	-g GAP		set extra gap between letters [default %d]\n\
 			valid GAP values are between %d and %d\n\
 	-e		don't echo sending to stdout [default echo]\n\
@@ -173,11 +187,14 @@ print_help ()
 	-o		don't allow [...] combinations [default combo]\n\
 	-p		don't allow {...} comments [default comments]\n\
 	-h		print this message\n\
-	-v		output version information and exit\n\n",
+	-V		output version information and exit\n\n",
 #endif /* not HAVE_GETOPT_LONG */
 	argv0,
+	cw_get_soundcard_file (),		cw_get_soundmixer_file (),
+	cw_get_console_file (),
 	cw_get_send_speed (),	min_speed,	max_speed,
 	cw_get_frequency (),	min_frequency,	max_frequency,
+	cw_get_volume (),	min_volume,	max_volume,
 	cw_get_gap (),		min_gap,	max_gap);
 	exit (0);
 }
@@ -193,19 +210,23 @@ static void
 parse_cmdline (int argc, char **argv)
 {
 	int	c;				/* Option character */
-	int	sound_fd;			/* Output file descriptor */
+	int	csound = FALSE, ssound = TRUE;	/* Sound source options */
+	char	*cdevice = NULL, *sdevice = NULL,
+		*mdevice = NULL;		/* Sound generation options */
 	int	intarg;				/* General integer arg holder */
 	int	argind;				/* Loop index */
 	char	env_options[ MAXOPTSTR ];	/* Env options string */
-	char	*sptr;				/* String pointer */
 	char	*local_argv[ MAXARGS ];		/* Local argv array */
-	int	local_argc = 0;			/* Local argc */
-	char	device[ MAXPATHLEN ];		/* Path to device file */
+	int	local_argc;			/* Local argc */
 #if defined(HAVE_GETOPT_LONG)
 	int	option_index;			/* Option index */
 	static const struct option long_options[] = {	/* Options table */
-		{ "device",	1, 0, 'd' }, { "file",	1, 0, 'f' },
-		{ "tone",	1, 0, 't' }, { "hz",	1, 0, 't' },
+		{ "sound",	1, 0, 's' },
+		{ "cdevice",	1, 0, 'd' },
+		{ "sdevice",	1, 0, 'x' },
+		{ "mdevice",	1, 0, 'y' }, { "file",	1, 0, 'f' },
+		{ "tone",	1, 0, 't' },
+		{ "volume",	1, 0, 'v' }, { "hz",	1, 0, 't' },
 		{ "wpm",	1, 0, 'w' }, { "gap",	1, 0, 'g' },
 		{ "noecho",	0, 0, 'e' }, { "nomessages",
 						     	0, 0, 'm' },
@@ -213,7 +234,7 @@ parse_cmdline (int argc, char **argv)
 		{ "nocombinations",
 				0, 0, 'o' },
 		{ "nocomments",	0, 0, 'p' }, { "help",	0, 0, 'h' },
-		{ "version",	0, 0, 'v' },
+		{ "version",	0, 0, 'V' },
 		{ 0, 0, 0, 0 }};
 #endif /* HAVE_GETOPT_LONG */
 
@@ -229,9 +250,12 @@ parse_cmdline (int argc, char **argv)
 	 * the strings from ..._OPTIONS, if defined, then putting the
 	 * command line args on (so they take precedence).
 	 */
+	local_argc = 0;
 	local_argv[ local_argc++ ] = argv[0];
 	if (getenv (ENV_OPTIONS) != NULL)
 	    {
+		char	*sptr;
+
 		strcpy (env_options, getenv (ENV_OPTIONS));
 		sptr = env_options;
 		while (local_argc < MAXARGS - 1)
@@ -261,42 +285,57 @@ parse_cmdline (int argc, char **argv)
 		local_argv[ local_argc++ ] = argv[ argind ];
 	    }
 
-	/* Initialize the sound device to stdout. */
-	sound_fd = dup (INITIAL_SOUNDFD);
-	if (sound_fd == -1)
-	    {
-		perror ("dup");
-		exit (1);
-	    }
-	strcpy (device, "stdout");
-
 	/* Process every option. */
 	while (TRUE)
 	    {
 #if defined(HAVE_GETOPT_LONG)
-		c = getopt_long (local_argc, local_argv, "d:f:t:w:g:emcophv",
+		c = getopt_long (local_argc, local_argv,
+					"s:d:x:y:f:t:v:w:g:emcophV",
 					long_options, &option_index);
 #else  /* not HAVE_GETOPT_LONG */
-		c = getopt (local_argc, local_argv, "d:f:t:w:g:emcophv");
+		c = getopt (local_argc, local_argv,
+					"s:d:x:y:f:t:v:w:g:emcophV");
 #endif /* not HAVE_GETOPT_LONG */
 		if (c == -1)
 			break;
 
 		switch (c)
 		    {
-			case 'd':
-				close (sound_fd);
-
-				sound_fd = open (optarg, O_RDWR);
-				if (sound_fd == -1)
+			case 's':
+				if (!strcmp (optarg, "console")
+						|| !strcmp (optarg, "c"))
+				    {
+					csound = TRUE; ssound = FALSE;
+				    }
+				else if (!strcmp (optarg, "soundcard")
+						|| !strcmp (optarg, "s"))
+				    {
+					csound = FALSE; ssound = TRUE;
+				    }
+				else if (!strcmp (optarg, "both")
+						|| !strcmp (optarg, "b"))
+				    {
+					csound = TRUE; ssound = TRUE;
+				    }
+				else
 				    {
 					fprintf (stderr,
-					    "%s: error opening output device\n",
+					    "%s: invalid sound source\n",
 									argv0);
-					perror (optarg);
 					exit (1);
 				    }
-				strcpy (device, optarg);
+				break;
+
+			case 'd':
+				cdevice = optarg;
+				break;
+
+			case 'x':
+				sdevice = optarg;
+				break;
+
+			case 'y':
+				mdevice = optarg;
 				break;
 
 			case 'f':
@@ -320,6 +359,16 @@ parse_cmdline (int argc, char **argv)
 				    }
 				break;
 
+			case 'v':
+				if (sscanf (optarg, "%d", &intarg) != 1
+					|| cw_set_volume (intarg) != 0 )
+				    {
+					fprintf (stderr,
+					   "%s: invalid volume value\n", argv0);
+					exit (1);
+				    }
+				break;
+
 			case 'w':
 				if (sscanf (optarg, "%d", &intarg) != 1
 					|| cw_set_send_speed (intarg) != 0 )
@@ -364,7 +413,7 @@ parse_cmdline (int argc, char **argv)
 				print_help ();
 				exit (0);
 
-			case 'v':
+			case 'V':
 				printf ("%s, ", VERSION);
 				printf ("%s\n", COPYRIGHT);
 				exit (0);
@@ -382,13 +431,54 @@ parse_cmdline (int argc, char **argv)
 	if (optind != local_argc)
 		print_usage ();
 
-	/* Check that the output device, default or selected, does sound. */
-	if (cw_set_file_descriptor (sound_fd) != 0)
+	/* Deal with odd argument combinations. */
+	if (!csound && cdevice != NULL)
 	    {
-		fprintf (stderr, "%s: output device won't do sound\n", argv0);
-		perror (device);
-		exit (1);
+		fprintf (stderr, "%s: no console sound: -d invalid\n", argv0);
+		print_usage ();
 	    }
+	if (!ssound && sdevice != NULL)
+	    {
+		fprintf (stderr, "%s: no soundcard sound: -x invalid\n", argv0);
+		print_usage ();
+	    }
+	if (!ssound && mdevice != NULL)
+	    {
+		fprintf (stderr, "%s: no soundcard sound: -y invalid\n", argv0);
+		print_usage ();
+	    }
+
+	/* First set up soundcard sound if required. */
+	if (ssound)
+	    {
+		if (sdevice != NULL)
+			cw_set_soundcard_file (sdevice);
+		if (cw_soundcard_possible () != 0)
+		    {
+			fprintf (stderr,
+				"%s: cannot set up soundcard sound\n", argv0);
+			perror (cw_get_soundcard_file ());
+			exit (1);
+		    }
+		if (mdevice != NULL)
+			cw_set_soundmixer_file (mdevice);
+	    }
+	cw_set_soundcard_sound (ssound);
+
+	/* Now set up console sound, again if required. */
+	if (csound)
+	    {
+		if (cdevice != NULL)
+			cw_set_console_file (cdevice);
+		if (cw_console_possible () != 0)
+		    {
+			fprintf (stderr,
+				"%s: cannot set up console sound\n", argv0);
+			perror (cw_get_console_file ());
+			exit (1);
+		    }
+	    }
+	cw_set_console_sound (csound);
 }
 
 
@@ -457,9 +547,9 @@ fprintf_cw (const char *format, ...)
 		cw_flush_tone_queue ();
 		exit (1);
 	    }
-	if (cw_tone_queue_wait () != 0)
+	if (cw_tone_queue_wait_critical (1) != 0)
 	    {
-		perror ("cw_tone_queue_wait");
+		perror ("cw_tone_queue_wait_critical");
 		cw_flush_tone_queue ();
 		exit (1);
 	    }
@@ -492,6 +582,11 @@ handle_stream_query (FILE *stream)
 				("%c%c%d", CW_STATUS_OK, c,
 				 			cw_get_frequency ());
 			break;
+		case CW_CMDV_VOLUME:
+			fprintf_stderr
+				("%c%c%d", CW_STATUS_OK, c,
+				 			cw_get_volume ());
+			break;
 		case CW_CMDV_SPEED:
 			fprintf_stderr
 				("%c%c%d", CW_STATUS_OK, c,
@@ -550,6 +645,10 @@ handle_stream_cwquery (FILE *stream)
 			fprintf_cw
 				("%d HZ ", cw_get_frequency ());
 			break;
+		case CW_CMDV_VOLUME:
+			fprintf_cw
+				("%d PERCENT ", cw_get_volume ());
+			break;
 		case CW_CMDV_SPEED:
 			fprintf_cw
 				("%d WPM ", cw_get_send_speed ());
@@ -592,7 +691,7 @@ handle_stream_cwquery (FILE *stream)
  * in, so this routine needs to read nothing from the input stream.
  */
 static void
-handle_stream_parameter (FILE *stream, unsigned int c, int new_value)
+handle_stream_parameter (unsigned int c, int new_value)
 {
 	/* Handle the new value according to the command used. */
 	switch (c)
@@ -608,6 +707,13 @@ handle_stream_parameter (FILE *stream, unsigned int c, int new_value)
 				return;
 			    }
 			break;
+		case CW_CMDV_VOLUME:
+			if (cw_set_volume (new_value) != 0)
+			    {
+				fprintf_stderr ("%c%c", CW_STATUS_ERR, c);
+				return;
+			    }
+			break;
 		case CW_CMDV_SPEED:
 			if (cw_set_send_speed (new_value) != 0)
 			    {
@@ -668,6 +774,7 @@ handle_stream_command (FILE *stream)
 			break;
 
 		case CW_CMDV_FREQUENCY:
+		case CW_CMDV_VOLUME:
 		case CW_CMDV_SPEED:
 		case CW_CMDV_GAP:
 		case CW_CMDV_ECHO:
@@ -680,7 +787,7 @@ handle_stream_command (FILE *stream)
 				fprintf_stderr ("%c%c", CW_STATUS_ERR, c);
 				return;
 			    }
-			handle_stream_parameter (stream, c, new_value);
+			handle_stream_parameter (c, new_value);
 			break;
 
 		case CW_CMD_QUERY:
@@ -741,9 +848,9 @@ handle_character_send (unsigned int c, int partial)
 	fprintf_stdout ("%c", c);
 
 	/* Wait for the character to complete. */
-	if (cw_tone_queue_wait () == -1)
+	if (cw_tone_queue_wait_critical (1) != 0)
 	    {
-		perror ("cw_tone_queue_wait");
+		perror ("cw_tone_queue_wait_critical");
 		cw_flush_tone_queue ();
 		exit (1);
 	    }
@@ -768,7 +875,7 @@ parse_stream (FILE *stream)
 	    {
 		/* Get a character, and check for end of file. */
 		c = getc (stream);
-		if (c == EOF)
+		if (c == (unsigned int) EOF)
 			return;
 
 		/*
@@ -872,8 +979,11 @@ parse_stream (FILE *stream)
  * Signal handler for signals, to stop tone before exit.
  */
 static void
-signal_handler (int sig)
+signal_handler (int signal_number)
 {
+	/* Avoid compiler warnings about unused argument. */
+	signal_number = 0;
+
 	/* Stop any tone and exit. */
 	cw_flush_tone_queue ();
 	exit (0);
@@ -911,6 +1021,7 @@ main (int argc, char **argv)
 
 	/* Read data from stream until end of file. */
 	parse_stream (stdin);
+	cw_tone_queue_wait ();
 
 	/* All done. */
 	return 0;
