@@ -1,6 +1,6 @@
 /*
  * UnixCW - Unix CW (Morse code) training program
- * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
+ * Copyright (C) 2001, 2002  Simon Baldwin (simonb@caldera.com)
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -17,7 +17,7 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *
  *
- * cwcp - Curses-based morse practice program, with menuing interface.
+ * cwcp - Curses-based Morse practice program, with menuing interface.
  *
  */
 
@@ -58,14 +58,25 @@
 #include "cwcpwords.h"
 
 /* Definitions. */
-#define	TITLE		"UNIX Morse Tutor V2.0, (C) 1997,2001 G0FRD"
-#define	VERSION		"cwcp version 2.0"
-#define	COPYRIGHT	"Copyright (C) 2001  Simon Baldwin\n"		\
+#define	TITLE		"UNIX Morse Tutor V2.2, (C) 1997-2002 Simon Baldwin"
+#define	VERSION		"cwcp version 2.2"
+#define	COPYRIGHT	"Copyright (C) 2001, 2002  Simon Baldwin\n"	\
 "This program comes with ABSOLUTELY NO WARRANTY; for details\n"		\
 "please see the file 'COPYING' supplied with the source code.\n"	\
 "This is free software, and you are welcome to redistribute it\n"	\
 "under certain conditions; again, see 'COPYING' for details.\n"		\
 "This program is released under the GNU General Public License."
+#define	INTRODUCTION							\
+"Cwcp is an interactive Morse code tutor program, designed\n"		\
+"both for learning Morse code for the first time, and for\n"		\
+"experienced Morse users who want, or need, to improve\n"		\
+"their receiving speed.\n\n"						\
+"To use the program, select a mode from those listed on\n"		\
+"the left, and begin sending by pressing Return or F9.\n\n"		\
+"You can vary the speed, tone, volume, and spacing of the\n"		\
+"Morse code at any time using the appropriate keys.\n\n"		\
+"To stop sending, press F9.  To stop the program, select\n"		\
+"Exit from the Mode menu, or use F12 or ^C.\n"
 
 #define	ASC_NUL			'\0'		/* End of string */
 #define	ASC_SPACE		' '		/* ASCII space char */
@@ -98,10 +109,13 @@
 #define	PSEUDO_KEYF10	('S'-CTRL_OFFSET)	/* Alternative FKEY10 */
 #define	PSEUDO_KEYF11	('D'-CTRL_OFFSET)	/* Alternative FKEY11 */
 #define	PSEUDO_KEYF12	('F'-CTRL_OFFSET)	/* Alternative FKEY12 */
+#define	PSEUDO_KEYNPAGE	('O'-CTRL_OFFSET)	/* Alternative PageDown */
+#define	PSEUDO_KEYPPAGE	('P'-CTRL_OFFSET)	/* Alternative PageUp */
 
 /* Step values for the UI control of the CW parameters. */
 #define	STEP_WPM		1
 #define	STEP_HZ			100
+#define	STEP_VOL		5
 #define	STEP_GAP		1
 #define	STEP_TIME		1
 #define	MIN_TIME		1
@@ -111,7 +125,8 @@
 static const short	colour_array[] = {
 		COLOR_BLACK, COLOR_RED, COLOR_GREEN, COLOR_YELLOW,
 		COLOR_BLUE, COLOR_MAGENTA, COLOR_CYAN, COLOR_WHITE };
-#define	NUM_COLOURS		(sizeof(colour_array)/sizeof(colour_array[0]))
+#define	NUM_COLOURS		((int) (sizeof (colour_array)	\
+					/ sizeof (colour_array[0])))
 #define	BOX_COLOURS		1		/* Normal colour pair */
 #define	DISP_COLOURS		2		/* Blue colour pair */
 #define	DISP_FGIND		7		/* White foreground */
@@ -136,7 +151,8 @@ static const char	*cwcp_kxffrp[]		= { CWCP_KXffRP };
 static const char	*cwcp_flyqc[]		= { CWCP_FLYQC };
 static const char	*cwcp_wj1gz[]		= { CWCP_WJ1GZ };
 static const char	*cwcp_x23789[]		= { CWCP_x23789 };
-static const char	*cwcp_ffffff[]		= { CWCP_ffffff };
+static const char	*cwcp_ffffff_1[]	= { CWCP_ffffff_1 };
+static const char	*cwcp_ffffff_2[]	= { CWCP_ffffff_2 };
 static const char	*cwcp_words[]		= { CWCP_3_LETTER_WORDS,
 						    CWCP_4_LETTER_WORDS,
 						    CWCP_5_LETTER_WORDS };
@@ -159,7 +175,7 @@ typedef	struct {
 	const char		**wordlist;	/* Word list if dictionary */
 	const int		list_length;	/* Length of word list */
 } smode_t;
-#define	DICTIONARY(A)		&(A)[0], (sizeof(A)/sizeof((A)[0]))
+#define	DICTIONARY(A)		&(A)[0], (sizeof (A) / sizeof ((A)[0]))
 static const smode_t	modes[] = {
 /*TOP*/	{ NULL,			M_NULL,		0,	NULL,	0 },
 	{ "Letter Groups     ",	M_DICTIONARY,	5,
@@ -193,13 +209,15 @@ static const smode_t	modes[] = {
 	{ "23789 Groups      ",	M_DICTIONARY,	5,
 					DICTIONARY (cwcp_x23789) },
 	{ ",?.;)/ Groups     ",	M_DICTIONARY,	5,
-					DICTIONARY (cwcp_ffffff) },
+					DICTIONARY (cwcp_ffffff_1) },
+	{ "\"'$(+:_ Groups    ",M_DICTIONARY,	5,
+					DICTIONARY (cwcp_ffffff_2) },
 	{ "Keyboard          ",	M_KEYBOARD,	0,	NULL,	0 },
-	{ "Exit              ",	M_EXIT,		0,	NULL,	0 },
+	{ "Exit(F12)         ",	M_EXIT,		0,	NULL,	0 },
 /*END*/	{ NULL,			M_NULL,		0,	NULL,	0 } };
 
 static int	current_mode	= 1;		/* Current program mode index */
-#define	NUM_MODES		((sizeof(modes)/sizeof(modes[0]))-2)
+#define	NUM_MODES		((sizeof (modes) / sizeof (modes[0])) - 2)
 
 /*
  * Table of keyboard translations to create a handful of prosigs from some
@@ -225,7 +243,6 @@ static const prosig_t	prosigs[] = {
 
 /* Initial values for some variables. */
 #define	INITIAL_TIME		15
-#define	INITIAL_FD		1		/* Initial sound on stdout */
 
 /* A handful of defines to send '73' on exit */
 #define	x73_SPEED		30
@@ -266,10 +283,10 @@ static int	box_bgind	= BOX_BGIND;	/* Black background */
 
 /* Curses windows. */
 static WINDOW	*scr, *base_bkgd;
-static WINDOW	*titlebox,  *modebox,  *charbox,  *wpmbox,  *hzbox,  *gapbox,
-		*timebox;
-static WINDOW	*titledisp, *modedisp, *chardisp, *wpmdisp, *hzdisp, *gapdisp,
-		*timedisp;
+static WINDOW	*titlebox,  *modebox,  *charbox,  *wpmbox,  *hzbox,  *volbox,
+		*gapbox,  *timebox;
+static WINDOW	*titledisp, *modedisp, *chardisp, *wpmdisp, *hzdisp, *voldisp,
+		*gapdisp, *timedisp;
 
 /* Base name of the program, from argv[0]. */
 static char *argv0	= NULL;
@@ -290,7 +307,7 @@ print_usage ()
 		"Try '%s -h' for more information.\n",
 #endif /* not HAVE_GETOPT_LONG */
 		argv0);
-	exit( 1 );
+	exit (1);
 }
 
 /*
@@ -303,22 +320,30 @@ print_help ()
 {
 	int	min_speed, max_speed;		/* Cw library speed limits */
 	int	min_frequency, max_frequency;	/* Cw library frequncy limits */
+	int	min_volume, max_volume;		/* Cw library volume limits */
 	int	min_gap, max_gap;		/* Cw library gap limits */
 	assert (argv0 != NULL);
 
 	/* Read the cw library limits on its operating parameters. */
 	cw_get_speed_limits	(&min_speed, &max_speed);
 	cw_get_frequency_limits	(&min_frequency, &max_frequency);
+	cw_get_volume_limits	(&min_volume, &max_volume);
 	cw_get_gap_limits	(&min_gap, &max_gap);
 
 	printf (
 #if defined(HAVE_GETOPT_LONG)
 	"Usage: %s [options...]\n\n\
-	-d, --device=DEVICE	use DEVICE for sound ioctl [default stdout]\n\
+	-s, --sound=SOURCE	generate sound on SOURCE [default 'soundcard']\n\
+				one of 's[oundcard]', 'c[onsole]', or 'b[oth]'\n\
+	-x, --sdevice=SDEVICE	use SDEVICE for soundcard [default %s]\n\
+	-y, --mdevice=MDEVICE	use MDEVICE for sound mixer [default %s]\n\
+	-d, --cdevice=CDEVICE	use CDEVICE for sound ioctl [default %s]\n\
 	-w, --wpm=WPM		set initial words per minute [default %d]\n\
 				valid WPM values are between %d and %d\n\
 	-t, --hz,--tone=HZ	set initial tone to HZ [default %d]\n\
 				valid HZ values are between %d and %d\n\
+	-v, --volume=PERCENT	set initial volume to PERCENT [default %d]\n\
+				valid PERCENT values are between %d and %d\n\
 	-g, --gap=GAP		set extra gap between letters [default %d]\n\
 				valid GAP values are between %d and %d\n\
 	-p, --time=TIME		set initial practice time [default %d mins]\n\
@@ -328,14 +353,20 @@ print_help ()
 	-m, --mono		specify no colours [default colours]\n\
 	-q, --no73		specify no cheery '73' on exit [default on]\n\
 	-h, --help		print this message\n\
-	-v, --version		output version information and exit\n\n",
+	-V, --version		output version information and exit\n\n",
 #else /* not HAVE_GETOPT_LONG */
 	"Usage: %s [options...]\n\n\
-	-d		use DEVICE for sound ioctl [default stdout]\n\
+	-s		generate sound on SOURCE [default 'soundcard']\n\
+			one of 's[oundcard]', 'c[onsole]', or 'b[oth]'\n\
+	-x SDEVICE	use SDEVICE for soundcard [default %s]\n\
+	-y MDEVICE	use MDEVICE for sound mixer [default %s]\n\
+	-d CDEVICE	use CDEVICE for sound ioctl [default %s]\n\
 	-w WPM		set initial words per minute [default %d]\n\
 			valid WPM values are between %d and %d\n\
 	-t HZ		set initial tone to HZ [default %d]\n\
 			valid HZ values are between %d and %d\n\
+	-v PERCENT	set initial volume to PERCENT [default %d]\n\
+			valid PERCENT values are between %d and %d\n\
 	-g GAP		set extra gap between letters [default %d]\n\
 			valid GAP values are between %d and %d\n\
 	-p TIME		set initial practice time [default %d mins]\n\
@@ -345,11 +376,14 @@ print_help ()
 	-m		specify no colours [default colours]\n\
 	-q		specify no cheery '73' on exit [default on]\n\
 	-h		print this message\n\
-	-v		output version information and exit\n\n",
+	-V		output version information and exit\n\n",
 #endif /* not HAVE_GETOPT_LONG */
 	argv0,
+	cw_get_soundcard_file (),		cw_get_soundmixer_file (),
+	cw_get_console_file (),
 	cw_get_send_speed (),	min_speed,	max_speed,
 	cw_get_frequency (),	min_frequency,	max_frequency,
+	cw_get_volume (),	min_volume,	max_volume,
 	cw_get_gap (),		min_gap,	max_gap,
 	INITIAL_TIME,		MIN_TIME,	MAX_TIME,
 	DISP_FGIND, DISP_BGIND, BOX_FGIND, BOX_BGIND);
@@ -366,27 +400,31 @@ static void
 parse_cmdline (int argc, char **argv)
 {
 	int	c;				/* Option character */
-	int	sound_fd = INITIAL_FD;		/* Output file descriptor */
+	int	csound = FALSE, ssound = TRUE;	/* Sound source options */
+	char	*cdevice = NULL, *sdevice = NULL,
+		*mdevice = NULL;		/* Sound generation options */
 	int	intarg;				/* General integer argument */
 	int	argind;				/* Loop index */
 	char	env_options[ MAXOPTSTR ];	/* Env options string */
 	char	*sptr;				/* String pointer */
 	char	*local_argv[ MAXARGS ];		/* Local argv array */
-	int	local_argc = 0;			/* Local argc */
-	char	device[ MAXPATHLEN ];		/* Path to device file */
+	int	local_argc;			/* Local argc */
 #if defined(HAVE_GETOPT_LONG)
 	int	option_index;			/* Option index */
 	static const struct option long_options[] = {	/* Options table */
-		{ "device",	1, 0, 'd' }, { "tone",	1, 0, 't' },
+		{ "sound",	1, 0, 's' },
+		{ "cdevice",	1, 0, 'd' },
+		{ "sdevice",	1, 0, 'x' },
+		{ "mdevice",	1, 0, 'y' },
+		{ "volume",	1, 0, 'v' }, { "tone",	1, 0, 't' },
 		{ "hz",		1, 0, 't' }, { "wpm",	1, 0, 'w' },
 		{ "gap",	1, 0, 'g' }, { "time",	1, 0, 'p' },
 		{ "colours",	1, 0, 'c' }, { "colors",1, 0, 'c' },
 		{ "mono",	0, 0, 'm' }, { "no73",	0, 0, 'q' },
 		{ "help",	0, 0, 'h' },
-		{ "version",	0, 0, 'v' },
+		{ "version",	0, 0, 'V' },
 		{ 0, 0, 0, 0 }};
 #endif /* HAVE_GETOPT_LONG */
-	strcpy (device, "stdout");
 
 	/* Set argv0 to be the basename part of the program name. */
 	argv0 = argv[0] + strlen (argv[0]);
@@ -400,6 +438,7 @@ parse_cmdline (int argc, char **argv)
 	 * the strings from ..._OPTIONS, if defined, then putting the
 	 * command line args on (so they take precedence).
 	 */
+	local_argc = 0;
 	local_argv[ local_argc++ ] = argv[0];
 	if (getenv (ENV_OPTIONS) != NULL)
 	    {
@@ -436,34 +475,53 @@ parse_cmdline (int argc, char **argv)
 	while (TRUE)
 	    {
 #if defined(HAVE_GETOPT_LONG)
-		c = getopt_long (local_argc, local_argv, "d:t:w:g:p:c:mqhv",
+		c = getopt_long (local_argc, local_argv,
+					"s:d:x:y:t:v:w:g:p:c:mqhV",
 					long_options, &option_index);
 #else /* not HAVE_GETOPT_LONG */
-		c = getopt (local_argc, local_argv, "d:t:w:g:p:c:mqhv");
+		c = getopt (local_argc, local_argv,
+					"s:d:x:y:t:v:w:g:p:c:mqhV");
 #endif /* not HAVE_GETOPT_LONG */
 		if (c == -1)
 			break;
 
 		switch (c)
 		    {
+			case 's':
+				if (!strcmp (optarg, "console")
+						|| !strcmp (optarg, "c"))
+				    {
+					csound = TRUE; ssound = FALSE;
+				    }
+				else if (!strcmp (optarg, "soundcard")
+						|| !strcmp (optarg, "s"))
+				    {
+					csound = FALSE; ssound = TRUE;
+				    }
+				else if (!strcmp (optarg, "both")
+						|| !strcmp (optarg, "b"))
+				    {
+					csound = TRUE; ssound = TRUE;
+				    }
+				else
+				    {
+					fprintf (stderr,
+					    "%s: invalid sound source\n",
+									argv0);
+					exit (1);
+				    }
+				break;
 
 			case 'd':
-				if (sscanf (optarg, "%s", device) != 1)
-				    {
-					fprintf (stderr,
-					    "%s: invalid device value\n",
-									argv0 );
-					exit (1);
-				    }
-				sound_fd = open (device, O_RDWR);
-				if (sound_fd == -1)
-				    {
-					fprintf (stderr,
-					    "%s: error opening output device\n",
-									argv0);
-					perror (device);
-					exit (1);
-				    }
+				cdevice = optarg;
+				break;
+
+			case 'x':
+				sdevice = optarg;
+				break;
+
+			case 'y':
+				mdevice = optarg;
 				break;
 
 			case 't':
@@ -476,6 +534,16 @@ parse_cmdline (int argc, char **argv)
 				    }
 				break;
 
+			case 'v':
+				if (sscanf (optarg, "%d", &intarg) != 1
+					|| cw_set_volume (intarg) != 0 )
+				    {
+					fprintf (stderr,
+					   "%s: invalid volume value\n", argv0);
+					exit (1);
+				    }
+				break;
+
 			case 'w':
 				if (sscanf (optarg, "%d", &intarg) != 1
 					|| cw_set_send_speed (intarg) != 0)
@@ -508,9 +576,9 @@ parse_cmdline (int argc, char **argv)
 				break;
 
 			case 'c':
-				if ( sscanf( optarg, "%d,%d,%d,%d",
+				if (sscanf (optarg, "%d,%d,%d,%d",
 						&disp_fgind, &disp_bgind,
-						&box_fgind,  &box_bgind ) != 4
+						&box_fgind,  &box_bgind) != 4
 					|| disp_fgind < 0
 					|| disp_fgind >= NUM_COLOURS
 					|| disp_bgind < 0
@@ -539,7 +607,7 @@ parse_cmdline (int argc, char **argv)
 				print_help ();
 				exit (0);
 
-			case 'v':
+			case 'V':
 				printf ("%s, ", VERSION);
 				printf ("%s\n", COPYRIGHT);
 				exit (0);
@@ -558,13 +626,54 @@ parse_cmdline (int argc, char **argv)
 	if (optind != local_argc)
 		print_usage ();
 
-	/* Check that the output device, default or selected, does sound. */
-	if (cw_set_file_descriptor (sound_fd) != 0)
+	/* Deal with odd argument combinations. */
+	if (!csound && cdevice != NULL)
 	    {
-		fprintf (stderr, "%s: output device won't do sound\n", argv0);
-		perror (device);
-		exit (1);
+		fprintf (stderr, "%s: no console sound: -d invalid\n", argv0);
+		print_usage ();
 	    }
+	if (!ssound && sdevice != NULL)
+	    {
+		fprintf (stderr, "%s: no soundcard sound: -x invalid\n", argv0);
+		print_usage ();
+	    }
+	if (!ssound && mdevice != NULL)
+	    {
+		fprintf (stderr, "%s: no soundcard sound: -y invalid\n", argv0);
+		print_usage ();
+	    }
+
+	/* First set up soundcard sound if required. */
+	if (ssound)
+	    {
+		if (sdevice != NULL)
+			cw_set_soundcard_file (sdevice);
+		if (cw_soundcard_possible () != 0)
+		    {
+			fprintf (stderr,
+				"%s: cannot set up soundcard sound\n", argv0);
+			perror (cw_get_soundcard_file ());
+			exit (1);
+		    }
+		if (mdevice != NULL)
+			cw_set_soundmixer_file (mdevice);
+	    }
+	cw_set_soundcard_sound (ssound);
+
+	/* Now set up console sound, again if required. */
+	if (csound)
+	    {
+		if (cdevice != NULL)
+			cw_set_console_file (cdevice);
+		if (cw_console_possible () != 0)
+		    {
+			fprintf (stderr,
+				"%s: cannot set up console sound\n", argv0);
+			perror (cw_get_console_file ());
+			exit (1);
+		    }
+	    }
+	cw_set_console_sound (csound);
 }
 
 
@@ -669,7 +778,7 @@ bye_bye ()
 {
 	/* Ignore the call if we don't want to send '73'. */
 	if (do_cheerio)
-	{
+	    {
 		/*
 		 * Set the sending parameters so that we get a speedy exit
 		 * from the program.  Then send the string.
@@ -683,7 +792,7 @@ bye_bye ()
 			exit (1);
 		    }
 		cw_tone_queue_wait ();
-	}
+	    }
 }
 
 
@@ -694,7 +803,7 @@ bye_bye ()
  * character, if one is ready, from the send queue and sends it.
  */
 static void
-dequeue_character()
+dequeue_character ()
 {
 	unsigned char		c;		/* Character from the queue */
 	int			max_y, max_x;	/* Window dimensions */
@@ -763,7 +872,7 @@ dequeue_character()
 				    }
 				if (cw_send_character (*sptr) != 0)
 				    {
-					perror ("cw_send_character_partial");
+					perror ("cw_send_character");
 					exit (1);
 				    }
 			    }
@@ -965,51 +1074,67 @@ interface_init ()
 	scrollok (chardisp, TRUE);
 	wrefresh (chardisp);
 
+	/* Add a bit of intro text to the main display. */
+	waddstr (chardisp, VERSION);
+	waddstr (chardisp, "\n\n");
+	waddstr (chardisp, INTRODUCTION);
+
 	/* Create the control feedback boxes. */
-	wpmbox = newwin (3, 20, 21, 0);
+	wpmbox = newwin (3, 16, 21, 0);
 	UI_INIT_BOX (wpmbox)
 	mvwaddstr (wpmbox, 0, 1, "Speed(F1-,F2+)");
 	wrefresh (wpmbox);
 
-	wpmdisp = newwin (1, 18, 22, 1);
+	wpmdisp = newwin (1, 14, 22, 1);
 	UI_INIT_DISPLAY (wpmdisp)
 	sprintf (buffer, "%2d WPM", cw_get_send_speed ());
-	mvwaddstr (wpmdisp, 0, 6, buffer);
+	mvwaddstr (wpmdisp, 0, 4, buffer);
 	wrefresh (wpmdisp);
 
-	hzbox = newwin (3, 20, 21, 20);
+	hzbox = newwin (3, 16, 21, 16);
 	UI_INIT_BOX (hzbox)
 	mvwaddstr (hzbox, 0, 1, "Tone(F3-,F4+)");
 	wrefresh (hzbox);
 
-	hzdisp = newwin (1, 18, 22, 21);
+	hzdisp = newwin (1, 14, 22, 17);
 	UI_INIT_DISPLAY (hzdisp)
 	sprintf (buffer, "%4d Hz", cw_get_frequency ());
-	mvwaddstr (hzdisp, 0, 5, buffer);
+	mvwaddstr (hzdisp, 0, 3, buffer);
 	wrefresh (hzdisp);
 
-	gapbox = newwin (3, 20, 21, 40);
+	volbox = newwin (3, 16, 21, 32);
+	UI_INIT_BOX (volbox)
+	mvwaddstr (volbox, 0, 1, "Vol(F5-,F6+)");
+	wrefresh (volbox);
+
+	voldisp = newwin (1, 14, 22, 33);
+	UI_INIT_DISPLAY (voldisp)
+	sprintf (buffer, "%3d %%", cw_get_volume ());
+	mvwaddstr (voldisp, 0, 4, buffer);
+	wrefresh (voldisp);
+
+	gapbox = newwin (3, 16, 21, 48);
 	UI_INIT_BOX (gapbox)
-	mvwaddstr (gapbox, 0, 1, "Spacing(F5-,F6+)");
+	mvwaddstr (gapbox, 0, 1, "Gap(F7-,F8+)");
 	wrefresh (gapbox);
 
-	gapdisp = newwin (1, 18, 22, 41);
+	gapdisp = newwin (1, 14, 22, 49);
 	UI_INIT_DISPLAY (gapdisp)
 	sprintf (buffer, "%2d dot%c",
 			cw_get_gap (), cw_get_gap () != 1 ? 's' : ' ');
-	mvwaddstr (gapdisp, 0, 5, buffer);
+	mvwaddstr (gapdisp, 0, 3, buffer);
 	wrefresh (gapdisp);
 
-	timebox = newwin (3, 20, 21, 60);
+	timebox = newwin (3, 16, 21, 64);
 	UI_INIT_BOX (timebox)
-	mvwaddstr (timebox, 0, 1, "Time(F7-,F8+)");
+	mvwaddstr (timebox, 0, 1, "Time(Dn-,Up+)");
 	wrefresh (timebox);
 
-	timedisp = newwin (1, 18, 22, 61);
+	timedisp = newwin (1, 14, 22, 65);
 	UI_INIT_DISPLAY (timedisp)
 	sprintf (buffer, " 0/%2d min%c",
 			practice_time, practice_time != 1 ? 's' : ' ');
-	mvwaddstr (timedisp, 0, 4, buffer);
+	mvwaddstr (timedisp, 0, 2, buffer);
 	wrefresh (timedisp);
 
 	/* Set up curses input mode. */
@@ -1115,7 +1240,7 @@ colour_update:
 
 speed_update:
 			sprintf (buffer, "%2d WPM", cw_get_send_speed ());
-			mvwaddstr (wpmdisp, 0, 6, buffer);
+			mvwaddstr (wpmdisp, 0, 4, buffer);
 			wrefresh (wpmdisp);
 			break;
 
@@ -1138,21 +1263,40 @@ speed_update:
 
 frequency_update:
 			sprintf (buffer, "%4d Hz", cw_get_frequency ());
-			mvwaddstr (hzdisp, 0, 5, buffer);
+			mvwaddstr (hzdisp, 0, 3, buffer);
 			wrefresh (hzdisp);
 			break;
 
 
 		case KEY_F(5):
 		case PSEUDO_KEYF5:
-		case KEY_NPAGE:
-			if (cw_set_gap (cw_get_gap () - STEP_GAP) == 0)
-				goto gap_update;
+			if (cw_set_volume
+					(cw_get_volume () - STEP_VOL) == 0)
+				goto volume_update;
 			break;
 
 		case KEY_F(6):
 		case PSEUDO_KEYF6:
-		case KEY_PPAGE:
+			if (cw_set_volume
+					(cw_get_volume () + STEP_VOL) == 0)
+				goto volume_update;
+			break;
+
+volume_update:
+			sprintf (buffer, "%3d %%", cw_get_volume ());
+			mvwaddstr (voldisp, 0, 4, buffer);
+			wrefresh (voldisp);
+			break;
+
+
+		case KEY_F(7):
+		case PSEUDO_KEYF7:
+			if (cw_set_gap (cw_get_gap () - STEP_GAP) == 0)
+				goto gap_update;
+			break;
+
+		case KEY_F(8):
+		case PSEUDO_KEYF8:
 			if (cw_set_gap (cw_get_gap () + STEP_GAP) == 0)
 				goto gap_update;
 			break;
@@ -1160,13 +1304,13 @@ frequency_update:
 gap_update:
 			sprintf (buffer, "%2d dot%c",
 				cw_get_gap (), cw_get_gap () != 1 ? 's' : ' ');
-			mvwaddstr (gapdisp, 0, 5, buffer);
+			mvwaddstr (gapdisp, 0, 3, buffer);
 			wrefresh (gapdisp);
 			break;
 
 
-		case KEY_F(7):
-		case PSEUDO_KEYF7:
+		case KEY_NPAGE:
+		case PSEUDO_KEYNPAGE:
 			if (practice_time > MIN_TIME)
 			    {
 				practice_time -= STEP_TIME;
@@ -1174,8 +1318,8 @@ gap_update:
 			    }
 			break;
 
-		case KEY_F(8):
-		case PSEUDO_KEYF8:
+		case KEY_PPAGE:
+		case PSEUDO_KEYPPAGE:
 			if (practice_time < MAX_TIME)
 			    {
 				practice_time += STEP_TIME;
@@ -1186,7 +1330,7 @@ gap_update:
 time_update:
 			sprintf (buffer, "%2d min%c",
 				practice_time, practice_time != 1 ? 's' : ' ');
-			mvwaddstr (timedisp, 0, 7, buffer);
+			mvwaddstr (timedisp, 0, 5, buffer);
 			wrefresh (timedisp);
 			break;
 
@@ -1234,6 +1378,7 @@ mode_update:
 			    {
 				bye_bye ();
 				interface_destroy ();
+				cw_tone_queue_wait ();
 				exit (0);
 			    }
 			else
@@ -1253,6 +1398,7 @@ mode_update:
 			    {
 				bye_bye ();
 				interface_destroy ();
+				cw_tone_queue_wait ();
 				exit (0);
 			    }
 			else
@@ -1277,6 +1423,7 @@ mode_update:
 			cw_flush_tone_queue ();
 			bye_bye ();
 			interface_destroy ();
+			cw_tone_queue_wait ();
 			exit (0);
 
 		case ASC_CTRLC:
@@ -1284,6 +1431,7 @@ mode_update:
 			cq_tail = cq_head;
 			cw_flush_tone_queue ();
 			interface_destroy ();
+			cw_tone_queue_wait ();
 			exit (0);
 	    }
 
@@ -1328,7 +1476,7 @@ cwlib_callback_event ()
 			 */
 			minutes = (time (NULL) - practice_start_timestamp) / 60;
 			sprintf (buffer, "%2d", minutes);
-			mvwaddstr (timedisp, 0, 4, buffer);
+			mvwaddstr (timedisp, 0, 2, buffer);
 			wrefresh (timedisp);
 
 			/* Check the time, and stop if over practice time. */
@@ -1363,7 +1511,7 @@ cwlib_callback_event ()
 			    }
 
 			/* Now dequeue the next character into the cw sender. */
-			dequeue_character();
+			dequeue_character ();
 			break;
 
 		case M_KEYBOARD:
@@ -1371,7 +1519,7 @@ cwlib_callback_event ()
 			 * If we are interested in keyboard data just dequeue
 			 * whatever is in the character queue.
 			 */
-			dequeue_character();
+			dequeue_character ();
 			break;
 
 		default:
@@ -1431,13 +1579,13 @@ interface_event (unsigned int c)
  * Signal handler for signals, to clear up on kill.
  */
 static void
-signal_handler (int sig)
+signal_handler (int signal_number)
 {
 	/* Attempt to wrestle the screen back from curses. */
 	interface_destroy ();
 
 	/* Show the signal caught, and exit. */
-	fprintf (stderr, "\nCaught signal %d, exiting...\n", sig);
+	fprintf (stderr, "\nCaught signal %d, exiting...\n", signal_number);
 	exit (1);
 }
 
