@@ -1,6 +1,6 @@
 /*
  * UnixCW - Unix CW (Morse code) training program
- * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
+ * Copyright (C) 2001, 2002  Simon Baldwin (simonb@caldera.com)
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -44,7 +44,8 @@
 			"L","M","N","O","P","Q","R","S","T","U","V",	\
 			"W","X","Y","Z",				\
 			"0","1","2","3","4","5","6","7","8","9",	\
-			"=","-",",","?",".",";",")","/"
+			"=","-",",","?",".",";",")","/",		\
+			"\"","'","$","(","+",":","_"
 #define	CWCP_EISH5	"E","I","S","H","5"
 #define	CWCP_TMO0	"T","M","O","0"
 #define	CWCP_AUV4	"A","U","V","4"
@@ -53,7 +54,8 @@
 #define	CWCP_FLYQC	"F","L","Y","Q","C"
 #define	CWCP_WJ1GZ	"W","J","1","G","Z"
 #define	CWCP_x23789	"2","3","7","8","9"
-#define	CWCP_ffffff	",","?",".",";",")","/"
+#define	CWCP_ffffff_1	",","?",".",";",")","/"
+#define	CWCP_ffffff_2	"\"","'","$","(","+",":","_"
 #define CWCP_3_LETTER_WORDS						\
 			"ACT","ADD","AGE","AGO","AID","AIM","AIR",	\
 			"ALL","AND","ANY","APE","ARC","ARE","ARK",	\
