@@ -1,6 +1,6 @@
 /*
  * UnixCW - Unix CW (Morse code) training program
- * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
+ * Copyright (C) 2001, 2002  Simon Baldwin (simonb@caldera.com)
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -82,13 +82,24 @@ cw_self_test_admin ()
 	printf ("cwlib: cw_get/set_debug flags tests complete\n");
 
 	/*
-	 * If stdout cannot do sound, proceed with tests silently.
+	 * Test setting sound output selections.
 	 */
-	if (cw_set_file_descriptor (1) != 0)
+	if (cw_console_possible () == -1)
 	    {
-		printf ("cwlib: stdout cannot do sound, proceeding silently\n");
+		printf ("cwlib: console device cannot do sound\n");
+		perror (cw_get_console_file ());
 		cw_set_console_sound (0);
 	    }
+	else
+		cw_set_console_sound (1);
+	if (cw_soundcard_possible () == -1)
+	    {
+		printf ("cwlib: soundcard device unavailable\n");
+		perror (cw_get_console_file ());
+		cw_set_soundcard_sound (0);
+	    }
+	else
+		cw_set_soundcard_sound (1);
 
 	return failures;
 }
@@ -105,6 +116,8 @@ cw_self_test_limits ()
 	int		cw_max_speed;
 	int		cw_min_frequency;	/* parameter limit locations */
 	int		cw_max_frequency;
+	int		cw_min_volume;		/* parameter limit locations */
+	int		cw_max_volume;
 	int		cw_min_gap;		/* parameter limit locations */
 	int		cw_max_gap;
 	int		cw_min_tolerance;	/* parameter limit locations */
@@ -119,6 +132,9 @@ cw_self_test_limits ()
 	cw_get_frequency_limits (&cw_min_frequency, &cw_max_frequency);
 	printf ("cwlib: cw_get_frequency_limits=%d,%d\n",
 					cw_min_frequency, cw_max_frequency);
+	cw_get_volume_limits (&cw_min_volume, &cw_max_volume);
+	printf ("cwlib: cw_get_volume_limits=%d,%d\n",
+					cw_min_volume, cw_max_volume);
 	cw_get_gap_limits (&cw_min_gap, &cw_max_gap);
 	printf ("cwlib: cw_get_gap_limits=%d,%d\n",
 					cw_min_gap, cw_max_gap);
@@ -225,6 +241,33 @@ cw_self_test_ranges ()
 	    }
 	printf ("cwlib: cw_set/get_frequency tests complete\n");
 
+	cw_get_volume_limits (&cw_min, &cw_max);
+	errno = 0;
+	error = cw_set_volume (cw_min - 1);
+	if (error != -1 || errno != EINVAL)
+	    {
+		printf ("cwlib: ERROR: cw_set_volume(cw_min_volume-1)\n");
+		failures++;
+	    }
+	errno = 0;
+	error = cw_set_volume (cw_max + 1);
+	if (error != -1 || errno != EINVAL)
+	    {
+		printf ("cwlib: ERROR: cw_set_volume(cw_max_volume+1)\n");
+		failures++;
+	    }
+	for (index = cw_min; index <= cw_max; index++)
+	    {
+		cw_set_volume (index);
+		if (cw_get_volume () != index)
+		    {
+			printf ("cwlib: ERROR: cw_get/set_volume\n");
+			failures++;
+			break;
+		    }
+	    }
+	printf ("cwlib: cw_set/get_volume tests complete\n");
+
 	cw_get_gap_limits (&cw_min, &cw_max);
 	errno = 0;
 	error = cw_set_gap (cw_min - 1);
@@ -337,11 +380,12 @@ cw_self_test_simple_tones ()
 	 * Ensure we can generate a few simple tones, and wait for them
 	 * to end.
 	 */
-	if (cw_queue_tone (100000, 8000) != 0
-			|| cw_queue_tone (100000, 4000) != 0
-			|| cw_queue_tone (100000, 2000) != 0)
+	cw_set_volume (70);
+	if (cw_queue_tone (100000, 4000) != 0
+			|| cw_queue_tone (100000, 2000) != 0
+			|| cw_queue_tone (100000, 1000) != 0)
 	    {
-		printf ("cwlib: ERROR: cw_queue_tone(10000, 8000|4000|2000)\n");
+		printf ("cwlib: ERROR: cw_queue_tone(10000, 4000|2000|1000)\n");
 		failures++;
 	    }
 	if (cw_tone_wait () != 0
@@ -396,6 +440,7 @@ cw_self_test_complex_tones ()
 	 * it shouldn't with this amount of data, then pause until it isn't
 	 * so full.
 	 */
+	cw_set_volume (70);
 	cw_get_frequency_limits (&cw_min, &cw_max);
 	for (index = cw_min; index < cw_max; index += 100)
 	    {
@@ -460,6 +505,7 @@ cw_self_test_tone_queue ()
 	 * queue, that it looks full when it is, and that we can flush it all
 	 * again afterwards, and recover.
 	 */
+	cw_set_volume (70);
 	printf ("cwlib: cw_get_tone_queue_capacity=%d\n",
 						cw_get_tone_queue_capacity ());
 	printf ("cwlib: empty cw_get_tone_queue_length=%d\n",
@@ -487,6 +533,66 @@ cw_self_test_tone_queue ()
 }
 
 
+/*
+ * cw_self_test_volumes()
+ */
+static int
+cw_self_test_volumes ()
+{
+	int		failures = 0;		/* count of tests that fail */
+	int		index;			/* loop count */
+	int		cw_min;			/* parameter limit locations */
+	int		cw_max;
+
+	/*
+	 * Fill the queue with short tones, then check that we can move the
+	 * volume through it's entire range.  Flush the queue when complete.
+	 */
+	cw_get_volume_limits (&cw_min, &cw_max);
+	while (!cw_tone_queue_full ())
+		cw_queue_tone (100000, 800);
+	for (index = cw_max; index >= cw_min; index -= 10)
+	    {
+		cw_tone_wait ();
+		if (cw_set_volume (index) != 0)
+		    {
+			printf ("cwlib: ERROR: cw_set_volume()\n");
+			failures++;
+		       	break;
+		    }
+		if (cw_get_volume () != index)
+		    {
+			printf ("cwlib: ERROR: cw_get_volume()\n");
+			failures++;
+		       	break;
+		    }
+		cw_tone_wait ();
+	    }
+	for (index = cw_min; index <= cw_max; index += 10)
+	    {
+		cw_tone_wait ();
+		if (cw_set_volume (index) != 0)
+		    {
+			printf ("cwlib: ERROR: cw_set_volume()\n");
+			failures++;
+		       	break;
+		    }
+		if (cw_get_volume () != index)
+		    {
+			printf ("cwlib: ERROR: cw_get_volume()\n");
+			failures++;
+		       	break;
+		    }
+		cw_tone_wait ();
+	    }
+	cw_tone_wait ();
+	cw_flush_tone_queue ();
+	printf ("cwlib: cw_set/get_volume tests complete\n");
+
+	return failures;
+}
+
+
 /*
  * cw_self_test_lookups()
  */
@@ -549,7 +655,8 @@ cw_self_test_dot_dash ()
 	/*
 	 * Send basic dot and dash using the library primitives.
 	 */
-	cw_set_send_speed (30); cw_set_frequency (800); cw_set_gap (0);
+	cw_set_send_speed (30); cw_set_gap (0); cw_set_frequency (800);
+	cw_set_volume (70);
 	if (cw_send_dot () != 0)
 	    {
 		printf ("cwlib: ERROR: cw_send_dot()\n");
@@ -1136,6 +1243,7 @@ cw_self_test_keyer ()
 	 * alternating dots and dashes.
 	 */
 	cw_set_send_speed (30); cw_set_gap (0); cw_set_frequency (800);
+	cw_set_volume (70);
 	if (cw_keyer_paddle_event (TRUE,  FALSE) != 0)
 	    {
 		printf ("cwlib: ERROR: cw_keyer_paddle_event\n");
@@ -1214,6 +1322,7 @@ cw_self_test_straightkey ()
 	 * Finally, a nice simple set of tests.
 	 */
 	cw_set_send_speed (30); cw_set_gap (0); cw_set_frequency (800);
+	cw_set_volume (70);
 	for (index = 0; index < 10; index++)
 	    {
 		if (cw_straightkey_event (FALSE) != 0)
@@ -1286,7 +1395,8 @@ cw_self_test ()
 		cw_self_test_admin,		cw_self_test_limits,
 		cw_self_test_ranges,		cw_self_test_tone_parameters,
 		cw_self_test_simple_tones,	cw_self_test_complex_tones,
-		cw_self_test_tone_queue,	cw_self_test_lookups,
+		cw_self_test_tone_queue,	cw_self_test_volumes,
+		cw_self_test_lookups,
 		cw_self_test_dot_dash,		cw_self_test_representations,
 		cw_self_test_characters,	cw_self_test_full_send,
 		cw_self_test_fixed_receive,	cw_self_test_adaptive_receive,
@@ -1305,13 +1415,31 @@ cw_self_test ()
 		testset = atoi (testset_value);
 
 	/*
-	 * If admin test not selected, set silent mode if necessary here.
+	 * If admin test not selected, set console and soundcard sound
+	 * according to what appears to be available.  This will generate
+	 * sound from both the soundcard and the console.  To get just
+	 * console sound, unplug the speakers.  To get just soundcard
+	 * sound, run the tests with stdout redirected to a non-console
+	 * device.
 	 */
-	if (testset && !(testset & 1)
-			&& cw_set_file_descriptor (1) != 0)
+	if (testset && !(testset & 1))
 	    {
-		printf ("cwlib: stdout cannot do sound, proceeding silently\n");
-		cw_set_console_sound (0);
+		if (cw_console_possible () == -1)
+		    {
+			printf ("cwlib: console device cannot do sound, %s\n",
+						cw_get_console_file ());
+			cw_set_console_sound (0);
+		    }
+		else
+			cw_set_console_sound (1);
+		if (cw_soundcard_possible () == -1)
+		    {
+			printf ("cwlib: sound device unavailable, %s\n",
+						cw_get_soundcard_file ());
+			cw_set_soundcard_sound (0);
+		    }
+		else
+			cw_set_soundcard_sound (1);
 	    }
 
 	/*
