@@ -1,6 +1,6 @@
 /*
  * UnixCW - Unix CW (Morse code) training program
- * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
+ * Copyright (C) 2001, 2002  Simon Baldwin (simonb@caldera.com)
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -58,9 +58,9 @@
 #include "cwlib.h"
 
 /* Strings displayed in 'about' dialog. */
-const char	*about_caption	= "Xcwcp version 2.0";
+const char	*about_caption	= "Xcwcp version 2.2";
 const char	*about_text	= \
-"Xcwcp version 2.0  Copyright (C) 2001  Simon Baldwin\n\n"
+"Xcwcp version 2.2  Copyright (C) 2002  Simon Baldwin\n\n"
 "This program comes with ABSOLUTELY NO WARRANTY; for details\n"
 "please see the file 'COPYING' supplied with the source code.\n"
 "This is free software, and you are welcome to redistribute it\n"
@@ -82,18 +82,31 @@ const char	*mode_whatsthis = \
 "To key CW into Xcwcp for receive mode, use either the mouse or the "
 "keyboard.  On the mouse, the left and right buttons form an Iambic "
 "keyer, and the middle mouse button works as a straight key.<br><br>"
-"On the keyboard, use the Control and Shift keys for Iambic keyer "
-"control, and the Alt key as a straight key.";
+"On the keyboard, use the Left and Right cursor keys, or the Control "
+"and Shift keys, for Iambic keyer control, and the Up or Down cursor "
+"keys, or the Space, Enter, or Return keys, as a straight key.";
 const char	*speed_whatsthis = \
 "This controls the CW sending speed.  If you deselect adaptive "
 "receive speed, it also controls the CW receiving speed.";
 const char	*frequency_whatsthis = \
-"This sets the frequency of the CW tone that sounds on the system "
-"console.<br><br>"
+"This sets the frequency of the CW tone on the system sound card "
+"or console.<br><br>"
 "It affects both sent CW and receive sidetone.";
+const char	*volume_whatsthis = \
+"This sets the volume of the CW tone on the system sound card.  "
+"It is not possible to control console sound volume, so in this "
+"case, all values other than zero produce tones.<br><br>"
+"The volume control affects both sent CW and receive sidetone.";
 const char	*gap_whatsthis = \
 "This sets the \"Farnsworth\" gap used in sending CW.  This gap is an "
 "extra number of dit-length silences between CW characters.";
+const char	*display_whatsthis = \
+"This is the main display for Xcwcp.  The random CW characters that "
+"Xcwcp generates, any keyboard input you type, and the CW that you "
+"key into Xcwcp all appear here.<br><br>"
+"You can clear the display contents from the File menu.<br><br>"
+"The status bar shows the current character being sent, any character "
+"received, and other general error and Xcwcp status information.";
 
 /*
  * Import the bulk dictionary data into character arrays, so it can be
@@ -111,7 +124,8 @@ static const char	*xcwcp_kxffrp[]		= { CWCP_KXffRP };
 static const char	*xcwcp_flyqc[]		= { CWCP_FLYQC };
 static const char	*xcwcp_wj1gz[]		= { CWCP_WJ1GZ };
 static const char	*xcwcp_x23789[]		= { CWCP_x23789 };
-static const char	*xcwcp_ffffff[]		= { CWCP_ffffff };
+static const char	*xcwcp_ffffff_1[]	= { CWCP_ffffff_1 };
+static const char	*xcwcp_ffffff_2[]	= { CWCP_ffffff_2 };
 static const char	*xcwcp_words[]		= { CWCP_3_LETTER_WORDS,
 						    CWCP_4_LETTER_WORDS,
 						    CWCP_5_LETTER_WORDS };
@@ -133,7 +147,7 @@ typedef	struct {
 	const char		**wordlist;	/* Word list if dictionary */
 	const int		list_length;	/* Length of word list */
 } smode_t;
-#define	DICTIONARY(A)		&(A)[0], (sizeof(A)/sizeof((A)[0]))
+#define	DICTIONARY(A)		&(A)[0], (sizeof (A) / sizeof ((A)[0]))
 static const smode_t	modes[] = {
 	{ "Letter Groups",	 M_DICTIONARY,	5,
 					DICTIONARY (xcwcp_alphabetic) },
@@ -166,7 +180,9 @@ static const smode_t	modes[] = {
 	{ "23789 Groups",	 M_DICTIONARY,	5,
 					DICTIONARY (xcwcp_x23789) },
 	{ ",?.;)/ Groups",	 M_DICTIONARY,	5,
-					DICTIONARY (xcwcp_ffffff) },
+					DICTIONARY (xcwcp_ffffff_1) },
+	{ "\"'$(+:_ Groups",	 M_DICTIONARY,	5,
+					DICTIONARY (xcwcp_ffffff_2) },
 	{ "Send Keyboard CW",	 M_KEYBOARD,	0,	NULL,	0 },
 	{ "Receive Keyed CW",	 M_RECEIVE,	0,	NULL,	0 },
 	{ NULL,			 M_NULL,	0,	NULL,	0 } };
@@ -178,7 +194,7 @@ static const smode_t	modes[] = {
 typedef struct {
 	const char		character;	/* Prosig character */
 	const char		*prosig;	/* Prosig equivalent */
-	const bool		combination;	/* true if characters sound as
+	const bool		combination;	/* True if characters sound as
 						   one single item */
 } prosig_t;
 static const prosig_t	prosigs[] = {
@@ -207,7 +223,7 @@ XcwcpApplication
 XcwcpApplication::XcwcpApplication ()
 	: QMainWindow (0, "Xcwcp", WDestructiveClose)
 {
-	/* Create the toolbar, and the start/stop button. */
+	/* Create a toolbar, and the start/stop button. */
 	QToolBar *toolbar = new QToolBar (this, "Xcwcp Operations");
 	toolbar->setLabel (tr ("Xcwcp Operations"));
 
@@ -232,25 +248,24 @@ XcwcpApplication::XcwcpApplication ()
 	mode_combo = new QComboBox (FALSE, toolbar, "Mode");
 	for (const smode_t *modep = modes; modep->type != M_NULL; modep++)
 		mode_combo->insertItem (modep->description);
-	connect (mode_combo, SIGNAL (activated(int)),
-						SLOT (mode_change()));
+	connect (mode_combo, SIGNAL (activated (int)),
+						SLOT (mode_change ()));
 
 	/*
-	 * Add the speed, frequency, and gap spin boxes.  Connect each to a
-	 * value change function, so that we can immediately pass on changes
-	 * in these values to the CW library.
+	 * Add the speed, frequency, volume, and gap spin boxes.  Connect
+	 * each to a value change function, so that we can immediately pass
+	 * on changes in these values to the CW library.
 	 */
 	int	cwlib_min, cwlib_max;
 
-	toolbar->addSeparator ();
 	new QLabel ("Speed:", toolbar, "Speed Label" );
 	toolbar->addSeparator ();
 	cw_get_speed_limits (&cwlib_min, &cwlib_max);
 	speed_spin = new QSpinBox (cwlib_min, cwlib_max, 1, toolbar, "Speed");
 	speed_spin->setSuffix (" WPM");
-	speed_spin->setValue (cw_get_send_speed());
-	connect (speed_spin, SIGNAL (valueChanged(int)),
-						SLOT (speed_change()));
+	speed_spin->setValue (cw_get_send_speed ());
+	connect (speed_spin, SIGNAL (valueChanged (int)),
+						SLOT (speed_change ()));
 
 	toolbar->addSeparator ();
 	new QLabel ("Tone:", toolbar, "Frequency Label" );
@@ -259,9 +274,20 @@ XcwcpApplication::XcwcpApplication ()
 	frequency_spin = new QSpinBox (cwlib_min, cwlib_max, 100, toolbar,
 								"Frequency");
 	frequency_spin->setSuffix (" Hz");
-	frequency_spin->setValue (cw_get_frequency());
-	connect (frequency_spin, SIGNAL (valueChanged(int)),
-						SLOT (frequency_change()));
+	frequency_spin->setValue (cw_get_frequency ());
+	connect (frequency_spin, SIGNAL (valueChanged (int)),
+						SLOT (frequency_change ()));
+
+	toolbar->addSeparator ();
+	new QLabel ("Volume:", toolbar, "Volume Label" );
+	toolbar->addSeparator ();
+	cw_get_volume_limits (&cwlib_min, &cwlib_max);
+	volume_spin = new QSpinBox (cwlib_min, cwlib_max, 5, toolbar,
+								"Volume");
+	volume_spin->setSuffix (" %");
+	volume_spin->setValue (cw_get_volume ());
+	connect (volume_spin, SIGNAL (valueChanged (int)),
+						SLOT (volume_change ()));
 
 	toolbar->addSeparator ();
 	new QLabel ("Gap:", toolbar, "Gap Label" );
@@ -269,9 +295,9 @@ XcwcpApplication::XcwcpApplication ()
 	cw_get_gap_limits (&cwlib_min, &cwlib_max);
 	gap_spin = new QSpinBox (cwlib_min, cwlib_max, 1, toolbar, "Gap");
 	gap_spin->setSuffix (" dot(s)");
-	gap_spin->setValue (cw_get_gap());
-	connect (gap_spin, SIGNAL (valueChanged(int)),
-						SLOT (gap_change()));
+	gap_spin->setValue (cw_get_gap ());
+	connect (gap_spin, SIGNAL (valueChanged (int)),
+						SLOT (gap_change ()));
 	toolbar->addSeparator ();
 
 	/* Finally for the toolbar, add whatsthis. */
@@ -284,6 +310,7 @@ XcwcpApplication::XcwcpApplication ()
 	QWhatsThis::add (mode_combo, mode_whatsthis);
 	QWhatsThis::add (speed_spin, speed_whatsthis);
 	QWhatsThis::add (frequency_spin, frequency_whatsthis);
+	QWhatsThis::add (volume_spin, volume_whatsthis);
 	QWhatsThis::add (gap_spin, gap_whatsthis);
 
 	/* Create the file popup menu. */
@@ -313,14 +340,14 @@ XcwcpApplication::XcwcpApplication ()
 	options->insertItem (reverse_paddles);
 	curtis_mode_b = new QCheckBox ("Curtis Mode B Timing", this,
 						"Curtis Mode B Timing");
-	connect (curtis_mode_b, SIGNAL (toggled(bool)),
-					SLOT (curtis_mode_b_change()));
+	connect (curtis_mode_b, SIGNAL (toggled (bool)),
+					SLOT (curtis_mode_b_change ()));
 	options->insertItem (curtis_mode_b);
 	adaptive_receive = new QCheckBox ("Adaptive CW Receive Speed", this,
 						"Adaptive CW Receive Speed");
 	adaptive_receive->setChecked (TRUE);
-	connect (adaptive_receive, SIGNAL (toggled(bool)),
-					SLOT (adaptive_receive_change()));
+	connect (adaptive_receive, SIGNAL (toggled (bool)),
+					SLOT (adaptive_receive_change ()));
 	options->insertItem (adaptive_receive);
 
 	/* Create the help popup menu. */
@@ -328,7 +355,7 @@ XcwcpApplication::XcwcpApplication ()
 	menuBar ()->insertSeparator ();
 	menuBar ()->insertItem ("&Help", help);
 
-	help->insertItem( "&About", this, SLOT (about ()), Key_F1);
+	help->insertItem ("&About", this, SLOT (about ()), Key_F1);
 
 	/* Add the CW display widget, and complete the GUI initializations. */
 	display = new XcwcpMultiLineEdit (this, "Display");
@@ -336,9 +363,10 @@ XcwcpApplication::XcwcpApplication ()
 	display->setReadOnly (TRUE);
 	display->setWordWrap (XcwcpMultiLineEdit::WidgetWidth);
 	display->setWrapPolicy (XcwcpMultiLineEdit::Anywhere);
+	QWhatsThis::add (display, display_whatsthis);
 	setCentralWidget (display);
 	statusBar ()->message ("Ready");
-	resize (600, 300);
+	resize (720, 320);
 
 	/* Initialize our circular buffer, and buffer state. */
 	cq_head = cq_tail = 0;
@@ -489,7 +517,7 @@ XcwcpApplication::cwlib_callback_event ()
  * character, if one is ready, from the send queue and sends it.
  */
 void
-XcwcpApplication::dequeue_character()
+XcwcpApplication::dequeue_character ()
 {
 	/* If dequeue is idle, ignore the call. */
 	if (cq_dequeue_state == CQ_IDLE)
@@ -541,7 +569,7 @@ XcwcpApplication::dequeue_character()
 			    }
 			if (cw_send_character (*sptr) != 0)
 			    {
-				perror ("cw_send_character_partial");
+				perror ("cw_send_character");
 				exit (1);
 			    }
 		    }
@@ -896,9 +924,10 @@ XcwcpApplication::start ()
 	 * library.  Sort of.  We can do it by just calling the slots for
 	 * the GUI widgets directly.
 	 */
-	speed_change();
-	frequency_change();
-	gap_change();
+	speed_change ();
+	frequency_change ();
+	volume_change ();
+	gap_change ();
 	curtis_mode_b_change ();
 
 	/*
@@ -997,7 +1026,15 @@ XcwcpApplication::new_xcwcp ()
 void
 XcwcpApplication::clear ()
 {
-	display->clear ();
+	/* If active, protect the display against callback updates. */
+	if (cwlib_use_state == STATE_ACTIVE)
+	    {
+		cw_block_callback (TRUE);
+		display->clear ();
+		cw_block_callback (FALSE);
+	    }
+	else
+		display->clear ();
 }
 
 
@@ -1036,6 +1073,7 @@ XcwcpApplication::sync_speed ()
 /*
  * speed_change()
  * frequency_change()
+ * volume_change()
  * gap_change()
  *
  * Handle changes in the spin boxes for these CW parameters.  The only action
@@ -1076,6 +1114,18 @@ XcwcpApplication::frequency_change ()
 	    }
 }
 void
+XcwcpApplication::volume_change ()
+{
+	if (cwlib_use_state == STATE_ACTIVE)
+	    {
+		if (cw_set_volume (volume_spin->value ()) != 0)
+		    {
+			perror ("cw_set_volume");
+			exit (1);
+		    }
+	    }
+}
+void
 XcwcpApplication::gap_change ()
 {
 	if (cwlib_use_state == STATE_ACTIVE)
@@ -1294,11 +1344,24 @@ XcwcpApplication::key_event (QKeyEvent *event)
 		/*
 		 * Handle receive modes.
 		 *
-		 * If this is the Alt key, use as a straight key.  There are
-		 * two Alt keys, usually, on a keyboard, so if one wears out,
-		 * there's always the other one.
+		 * If this is a key press that it not the first one of
+		 * an autorepeating key, ignore the event.  This prevents
+		 * autorepeat from getting in the way of identifying the
+		 * real keyboard events we are after.
 		 */
-		if (event->key () == Key_Alt)
+		if (event->isAutoRepeat ())
+			return;
+
+		/*
+		 * If this is the Space, UpArrow, DownArrow, Enter, or
+		 * Return key, use as a straight key.  If one wears out,
+		 * there's always the other ones.
+		 */
+		if (event->key () == Key_Space
+				|| event->key () == Key_Up
+				|| event->key () == Key_Down
+				|| event->key () == Key_Enter
+				|| event->key () == Key_Return)
 		    {
 			if (event->type () == QEvent::KeyPress)
 				cw_straightkey_event (TRUE);
@@ -1309,10 +1372,12 @@ XcwcpApplication::key_event (QKeyEvent *event)
 		    }
 
 		/*
-		 * If this is the Ctrl key, use as one of the paddles.  Which
-		 * paddle depends on the reverse_paddles state.
+		 * If this is the Ctrl/LeftArrow key, use as one of the
+		 * paddles.  Which paddle depends on the reverse_paddles
+		 * state.
 		 */
-		if (event->key () == Key_Control)
+		if (event->key () == Key_Control
+				|| event->key () == Key_Left)
 		    {
 			if (event->type () == QEvent::KeyPress)
 			    {
@@ -1332,10 +1397,11 @@ XcwcpApplication::key_event (QKeyEvent *event)
 		    }
 
 		/*
-		 * If this is the Shift key, use as the other one of the
-		 * paddles.
+		 * If this is the Shift/RightArrow key, use as the other
+		 * one of the paddles.
 		 */
-		if (event->key () == Key_Shift)
+		if (event->key () == Key_Shift
+				|| event->key () == Key_Right)
 		    {
 			if (event->type () == QEvent::KeyPress)
 			    {
@@ -1381,32 +1447,60 @@ XcwcpApplication::mouse_event (QMouseEvent *event)
 	int mode = mode_combo->currentItem ();
 	if (modes[mode].type == M_RECEIVE)
 	    {
-		/* Map mouse buttons 1/2/3 to Ctrl/Alt/Shift keys. */
-		int	key = Key_unknown;
-		if (event->button () == LeftButton)
-			key = Key_Control;
-		else if (event->button () == MidButton)
-			key = Key_Alt;
-		else if (event->button () == RightButton)
-			key = Key_Shift;
-		else
-			return;
-
-		/*
-		 * Map mouse button presses and double clicks onto key
-		 * presses, and button releases onto key releases.
-		 */
-		QEvent::Type	key_event_type = QEvent::None;
-		if (event->type () == QEvent::MouseButtonDblClick
-				|| event->type () == QEvent::MouseButtonPress)
-			key_event_type = QEvent::KeyPress;
-		else
-			if (event->type () == QEvent::MouseButtonRelease)
-				key_event_type = QEvent::KeyRelease;
+		/* Use the middle mouse button as a straight key. */
+		if (event->button () == MidButton)
+		    {
+			if (event->type () == QEvent::MouseButtonPress
+					|| event->type ()
+						== QEvent::MouseButtonDblClick)
+				cw_straightkey_event (TRUE);
 			else
-				return;
+				if (event->type ()
+						== QEvent::MouseButtonRelease)
+					cw_straightkey_event (FALSE);
+			return;
+		    }
 
-		/* Call the key event handler with a constructed event. */
-		key_event (new QKeyEvent (key_event_type, key, 0, 0));
+		/* Use left and right mouse buttons as paddles. */
+		if (event->button () == LeftButton)
+		    {
+			if (event->type () == QEvent::MouseButtonPress
+					|| event->type ()
+						== QEvent::MouseButtonDblClick)
+			    {
+				if (!reverse_paddles->isChecked ())
+					cw_keyer_dot_paddle_event (TRUE);
+				else
+					cw_keyer_dash_paddle_event (TRUE);
+			    }
+			else if (event->type () == QEvent::MouseButtonRelease)
+			    {
+				if (!reverse_paddles->isChecked ())
+					cw_keyer_dot_paddle_event (FALSE);
+				else
+					cw_keyer_dash_paddle_event (FALSE);
+			    }
+			return;
+		    }
+		if (event->button () == RightButton)
+		    {
+			if (event->type () == QEvent::MouseButtonPress
+					|| event->type ()
+						== QEvent::MouseButtonDblClick)
+			    {
+				if (!reverse_paddles->isChecked ())
+					cw_keyer_dash_paddle_event (TRUE);
+				else
+					cw_keyer_dot_paddle_event (TRUE);
+			    }
+			else if (event->type () == QEvent::MouseButtonRelease)
+			    {
+				if (!reverse_paddles->isChecked ())
+					cw_keyer_dash_paddle_event (FALSE);
+				else
+					cw_keyer_dot_paddle_event (FALSE);
+			    }
+			return;
+		    }
 	    }
 }
