@@ -1,6 +1,6 @@
 /*
  * UnixCW - Unix CW (Morse code) training program
- * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
+ * Copyright (C) 2001, 2002  Simon Baldwin (simonb@caldera.com)
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -53,7 +53,8 @@
 #define	CWCP_FLYQC	"F","L","Y","Q","C"
 #define	CWCP_WJ1GZ	"W","J","1","G","Z"
 #define	CWCP_x23789	"2","3","7","8","9"
-#define	CWCP_ffffff	",","?",".",";",")","/"
+#define	CWCP_ffffff_1	",","?",".",";",")","/"
+#define	CWCP_ffffff_2	"\"","'","$","(","+",":","_"
 #define CWCP_3_LETTER_WORDS						\
 			"ACT","ADD","AGE","AGO","AID","AIM","AIR",	\
 			"ALL","AND","ANY","APE","ARC","ARE","ARK",	\
